/**
 * ADT that represents a 3D game object in the world.  It has a name,
 * a position, an orientation, and a size.
 *
 * A note on equality:  there are two types of equality, (1) equality in the
 * literal programatic sense where the member values are the same (i.e name,
 * position, orietation, and size), and (2) equality in the game logic sense
 * where the name serves as a unique identifier.  Here the name describes the
 * particular class of objects like "Ash Yam", while the GameObject itself
 * describes a specific instance, "The big upside down Ash Yam under that table".
 */
#ifndef _GAME_OBJECT_H
#define _GAME_OBJECT_H
#include "EsmSubCellRef.h"

typedef struct {
	float x, y, z;
} vector3d;

class GameObject {
	
	const TCHAR* name;
	// position is center of object.
	// rot is in radians.
	// size is distance from center (position) to edge.
	vector3d pos, rot, size;  
	

	
public:
	GameObject(const TCHAR* pName);
	GameObject(const TCHAR* pName, vector3d pos, vector3d rot, vector3d size);

	// getter methods
	const TCHAR* getName() { return name; }
	
	vector3d getPos() { return pos; }
	float getXpos() { return pos.x; }
	float getYpos() { return pos.y; }
	float getZpos() { return pos.z; }
	
	vector3d getRot() { return rot; }
	float getXrot() { return rot.x; }
	float getYrot() { return rot.y; }
	float getZrot() { return rot.z; }
	
	vector3d getSize() { return size; }
	float getXsize() { return size.x; }
	float getYsize() { return size.y; }
	float getZsize() { return size.z; }

	void setPos(vector3d pos) {
		this->pos.x = pos.x;
		this->pos.y = pos.y;
		this->pos.z = pos.z;
	}
	void setXpos(float x) { pos.x = x; }
	void setYpos(float y) { pos.y = y; }
	void setZpos(float z) { pos.z = z; }
	
	void setRot(vector3d rot) {
		this->rot.x = rot.x;
		this->rot.y = rot.y;
		this->rot.z = rot.z;
	}
	void setXrot(float x) { rot.x = x; }
	void setYrot(float y) { rot.y = y; }
	void setZrot(float z) { rot.z = z; }

	void setSize(vector3d size) {
		this->size.x = size.x;
		this->size.y = size.y;
		this->size.z = size.z;
	}
	void setXsize(float x) { size.x = x; }
	void setYsize(float y) { size.y = y; }
	void setZsize(float z) { size.z = z; }

	// synch game object positions and orientations
	void synchObjects(GameObject* obj) { setPos(obj->getPos()); setRot(obj->getRot()); }

	// placement methods
	void putAbove(GameObject* obj);
	void putBelow(GameObject* obj);
	void putLeft(GameObject* obj);
	void putRight(GameObject* obj);
	void putInFront(GameObject* obj);
	void putBehind(GameObject* obj);

	// equality methods
	
	// used for type equality for game logic related methods, like presence in a container
	bool sameName(GameObject* obj) { return (name == obj->getName()); }
	
	// used for object equality, hence is virtual and should be over ridden by subclassing methods
	virtual bool equals(GameObject* obj);

	// output methods
	CEsmSubCellRef* toCREF();
};

#endif