#include "graph.h"

void Graph::addNode(const char* node) {
	if (!hasNode(node)) {
		forward_map[node];
		backward_map[node];
	}
}

void Graph::removeNode(const char* node) {
	if (hasNode(node)) {
		forward_map.erase(node);
		backward_map.erase(node);
	}

	adjacency_map::const_iterator it;
	for (it = forward_map.begin(); it != forward_map.end(); it++) {
		removeEdge((*it).first, node);
	}
}

// if edge already exists, update its weight
void Graph::addEdge(const char* from, const char* to, int weight) {
	if (!hasEdge(from, to)) {
		if (!hasNode(from)) 
			addNode(from);
		
		if (!hasNode(to)) 
			addNode(to);
	}
	forward_map[from][to] = weight;
	backward_map[to][from] = weight;
}

void Graph::removeEdge(const char* from, const char* to) {
	if (hasEdge(from, to)) {
		forward_map[from].erase(to);
		backward_map[to].erase(from);
	}
}

bool Graph::hasNode(const char* node) const {
	return forward_map.count(node) != 0;
}

bool Graph::hasEdge(const char* from, const char* to) const {
	if (hasNode(from) && hasNode(to)) {
		return forward_map.find(from)->second.count(to) != 0;
	}
	return false;
}

bool Graph::hasPath(const char* from, const char* to, adjacency_map& amap) const {
	if (!hasNode(from) || !hasNode(to)) return false;
	if (strcmp(from, to) == 0) return true;			// the case where from == to
	
	std::set<const char*> visited;
	std::deque<const char*> q;
	
	// breadth first search for path
	q.push_back(from);
	do {
		const char* node = q.front();	
		q.pop_front();
		if (strcmp(node, to) == 0) return true;
		visited.insert(node);

		edge_map::const_iterator it;
		for (it = amap[node].begin(); it != amap[node].end(); it++) {
			if (visited.count(it->first) == 0) {
				q.push_front(it->first);
			}
		}
	} while(!q.empty());
	
	return false;
}

bool Graph::hasForwardPath(const char* from, const char* to) const {
	return hasPath(from, to, forward_map);
}

bool Graph::hasBackwardPath(const char* from, const char* to) const {
	return hasPath(from, to, backward_map);
}

int Graph::edgeWeight(const char* from, const char* to) const {
	if (hasEdge(from, to))
		return (*(*forward_map.find(from)).second.find(to)).second;
	return -1;
}

edge_map Graph::getAdjacent(const char* node) const {
	edge_map emap;
	if (hasNode(node)) {
		emap = (*forward_map.find(node)).second;
	}
	return emap;
}

edge_map Graph::getAdjacentInv(const char* node) const {
	edge_map emap;
	if (hasNode(node)) {
		emap = (*backward_map.find(node)).second;
	}
	return emap;
}

int Graph::inDegree(const char* node) const {
	if (hasNode(node)) {
		return (*backward_map.find(node)).second.size();
	}
	return -1;
}

int Graph::outDegree(const char* node) const {
	if (hasNode(node)) {
		return (*forward_map.find(node)).second.size();
	}
	return -1;
}

void Graph::print() const {
	adjacency_map::const_iterator a_iter;
	edge_map::const_iterator e_iter;

	for (a_iter = forward_map.begin(); a_iter != forward_map.end(); a_iter++) {
		edge_map emap = a_iter->second;
		if (emap.empty()) continue;
		for (e_iter = emap.begin(); e_iter != emap.end(); e_iter++) {
			printf("%s -> %s\n", a_iter->first, e_iter->first);
		}
	}

	for (a_iter = backward_map.begin(); a_iter != backward_map.end(); a_iter++) {
		edge_map emap = a_iter->second;
		for (e_iter = emap.begin(); e_iter != emap.end(); e_iter++) {
			printf("%s -> %s\n", a_iter->first, e_iter->first);
		}
	}

}