#include "room.h"

Room::Room(const Graph &ig, const Graph &pg) : floor("floor"), room("room"), igraph(ig), pgraph(pg) {
	// automatically add room, floor, etc...
	objectMap["floor"].push_back(&floor);
	floor.setZpos(-128);
	floor.setXsize(325);
	floor.setYsize(325);
	objectMap["room"].push_back(&room);
	//printf("%d %d\n", objectMap.count("floor"), objectMap.count("room"));
}

Room::~Room() {
	objectMap.erase("floor");
	objectMap.erase("room");
	game_obj_map::iterator gom_iter;
	game_obj_list::iterator gol_iter;

	for (gom_iter = objectMap.begin(); gom_iter != objectMap.end(); gom_iter++) {
		game_obj_list gol = gom_iter->second;
		for (gol_iter = gol.begin(); gol_iter != gol.end();   gol_iter++) {
			delete (*gol_iter);
		}
	}
}
bool Room::tryToPlace(const char* gameObj, const char* subj) {
	if (!igraph.hasNode(subj)) {
		printf("error: couldn't find is_a mapping for %s\n", subj);
		return false;
	}

	edge_map isa = igraph.getAdjacent(subj);
	edge_map::const_iterator isa_iter;

	for (isa_iter = isa.begin(); isa_iter != isa.end(); isa_iter++) {
		const char* nextSubj = (*isa_iter).first;
		
		if (!pgraph.hasNode(nextSubj)) {
			// not really an error/warning
			printf("warning: couldn't find pred obj list for %s\n", nextSubj);
			continue;		
		}

		edge_map pred = pgraph.getAdjacent(nextSubj);
		edge_map::const_iterator pred_iter;

		for (pred_iter = pred.begin(); pred_iter != pred.end(); pred_iter++) {
			std::pair<const char*, int> pop = (*pred_iter);
			if (addObject(gameObj, pop.second, pop.first)) {
				return true;
			}
		}

		// recurse down the [isA] heirarchy
		if (tryToPlace(gameObj, nextSubj)) {
			return true;
		}
	}
	return false;
}

// ignoring rotation for now, AABB (axis aligned bounding box intersecion test)
bool Room::overlaps(GameObject *subj, GameObject *obj) {
	float xdiff = subj->getXpos() - obj->getXpos();
	float ydiff = subj->getYpos() - obj->getYpos();
	float zdiff = subj->getZpos() - obj->getZpos();

	// ok for boundries to touch hence < and not <=
	return ( fabs(xdiff) < (subj->getXsize() + obj->getXsize() - .1) &&
			 fabs(ydiff) < (subj->getYsize() + obj->getYsize() - .1) &&
			 fabs(zdiff) < (subj->getZsize() + obj->getZsize() - .1));
}

bool Room::canPlace(GameObject* subj) {
	game_obj_map::const_iterator gom_iter;
	game_obj_list::iterator it;
	
	long index = 1;
	for (gom_iter = objectMap.begin(); gom_iter != objectMap.end(); gom_iter++) {
		if (strcmp(gom_iter->first, "floor") == 0 || 
			strcmp(gom_iter->first, "room") == 0) continue;
		
		game_obj_list objects = (*gom_iter).second;
		for (it = objects.begin(); it != objects.end(); it++) {
			if (overlaps(subj, (*it))) { 
				printf("%s overlaps %s\n", subj->getName(), (*it)->getName());
				return false;
			}
		}
	}
	return true;
}

// x, y positions match
bool Room::centerOf(GameObject* subj, GameObject* obj) {
	subj->setXpos(obj->getXpos());
	subj->setYpos(obj->getYpos());
	return true;
}

// match outer x,y w/ outer x,y, rotate so front faces in
bool Room::around(GameObject* subj, GameObject* obj, int angle) {
	subj->setYpos(obj->getYpos());

	if (angle == 0) {
		subj->setXpos(obj->getXpos() - obj->getXsize() - subj->getXsize());
		subj->setZrot(3*M_PI/2);	
	} else if (angle == 1) {
		subj->setXpos(obj->getXpos() + obj->getXsize() + subj->getXsize());
		subj->setZrot(M_PI/2);
	}

	return true;
}

// match outer x,y w/ inner x,y, rotate so front faces in
bool Room::edgeOf(GameObject* subj, GameObject* obj, int angle) {
	subj->setYpos(obj->getYpos());
	if (angle == 0) {
		subj->setXpos(obj->getXpos() - obj->getXsize() + subj->getXsize());
		subj->setZrot(3*M_PI/2);
	} else if (angle == 1) {
		subj->setXpos(obj->getXpos() + obj->getXsize() - subj->getXsize());
		subj->setZrot(M_PI/2);
	}
	
	return true;
}

// match corner w/ corner (inside)
bool Room::cornerOf(GameObject* subj, GameObject* obj) {
	int corner = 0;
	switch(corner) {
		case 0:
			subj->setYpos(obj->getYpos() + obj->getYsize() - subj->getYsize());
			subj->setXpos(obj->getXpos() - obj->getXsize() + subj->getXsize());
			break;
	}
	return true;
}

// subj lower z matches obj upper z
bool Room::onTopOf(GameObject* subj, GameObject* obj) {
	subj->setZpos(obj->getZpos() + obj->getZsize() + subj->getZsize());
	return true;
}

// match upper z w/ lower z
bool Room::below(GameObject* subj, GameObject* obj) {
	subj->setZpos(obj->getZpos() - obj->getZsize() - subj->getZsize());
	return true;
}

// match lower z's
bool Room::beside(GameObject* subj, GameObject* obj) {
	subj->setZpos((obj->getZpos() - obj->getZsize()) + subj->getZsize());
	return true;
}

bool Room::addObject(const char* subj, int predicate, const char* obj) {
	bool status = false;
	string predString;
	GameObject* gameObject = NULL;
	GameObject* gameSubject = NULL;
	
	std::set<const char*> visited;
	std::deque<const char*> q;
	
	// breadth first search of "is a" paths
	q.push_back(obj);
	do {
		const char* objectName = q.front();	
		q.pop_front();
		visited.insert(objectName);
		//printf("subj name = %s, obj name = %s %d\n", subj, objectName, objectMap.count(objectName));
		
		game_obj_list::const_iterator gol_iter;
		for (gol_iter = objectMap[objectName].begin(); gol_iter != objectMap[objectName].end(); gol_iter++) {
			gameObject = (*gol_iter);
			//printf("game obj = %s\n", gameObject->getName());
			gameSubject = new GameObject(_T(subj));
			predString = "";
			bool ard = false, edge = false;
			int angle = 0;

			if (strcmp(subj, "misc_com_redware_vase") == 0) {
				vector3d size = {15.403, 14.894, 23.605};
				gameSubject->setSize(size);
			} else if (strcmp(subj, "furn_com_r_table_01") == 0) {
				vector3d size = {96.671, 52.597, 39.723};
				gameSubject->setSize(size);
			} else if (strcmp(subj, "furn_com_r_chair_01") == 0) {
				vector3d size = {36.529, 35.410, 49.952};
				gameSubject->setSize(size);
			} else if (strcmp(subj, "misc_imp_silverware_plate_01") == 0) {
				vector3d size = {17.209, 17.209, 3.25};
				gameSubject->setSize(size);
			} else if (strcmp(subj, "ingred_ash_yam_01") == 0) {
				vector3d size = {9.854, 17.844, 5.314};
				gameSubject->setSize(size);
			} else if (strcmp(subj, "furn_com_r_bookshelf_01") == 0) {
				vector3d size = {84.288,28.781,98.871};
				gameSubject->setSize(size);
			}

			if ((predicate & CENTER_OF) == CENTER_OF) {
				predString += "<center of> ";
				centerOf(gameSubject, gameObject);
			} else if ((predicate & AROUND) == AROUND) {
				predString += "<around> ";
				ard = true;
				around(gameSubject, gameObject, angle);
			} else if ((predicate & EDGE_OF) == EDGE_OF) {
				predString += "<edge of> ";
				edge = true;
				edgeOf(gameSubject, gameObject, angle);
			} else if ((predicate & CORNER_OF) == CORNER_OF) {
				predString += "<corner of> ";
				cornerOf(gameSubject, gameObject);
			}

			if ((predicate & ON_TOP_OF) == ON_TOP_OF) {
				predString += "<on top of> ";
				onTopOf(gameSubject, gameObject);
			} else if ((predicate & BESIDE) == BESIDE) {
				predString += "<beside> ";
				beside(gameSubject, gameObject);
			}
			
			printf("trying to place: %s %s %s\n", subj, predString.c_str(), gameObject->getName());
			if (canPlace(gameSubject)) { 
				status = true;
				break;
			} else if (edge || ard) {
				while (!status && angle < 1) {
					angle++;
					if (edge) edgeOf(gameSubject, gameObject, angle);
					else if (ard) around(gameSubject, gameObject, angle);

					if (canPlace(gameSubject)) status = true;
				}
			}
		}

		edge_map emap = igraph.getAdjacentInv(objectName);
		edge_map::const_iterator it;
		for (it = emap.begin(); it != emap.end(); it++) {
			q.push_front(it->first);
			//printf("adding %s\n", it->first);
		}

	} while(!q.empty() && !status);

	if (status && gameSubject) { objectMap[subj].push_back(gameSubject); }
	else if (gameSubject != NULL) { delete gameSubject;}

	//printf("%f %f %f\n", gameSubject->getXpos(), gameSubject->getYpos(), gameSubject->getZpos());
	//printf("0x%x\n", gameSubject);

	return status;
}

// assuming there's no loops in the "is a" graph
const char* Room::hasObject(const char* obj) const {
	if (objectMap.count(obj) != 0) return obj;
	if (!igraph.hasNode(obj)) return 0;
	
	game_obj_map::const_iterator gom_iter;
	for (gom_iter = objectMap.begin(); gom_iter != objectMap.end(); gom_iter++) {
		if (igraph.hasBackwardPath(obj, gom_iter->first)) return gom_iter->first;
	}

	return 0;
}

void Room::toRecord(CEsmRecord* pRecord) const {
	game_obj_map::const_iterator gom_iter;
	game_obj_list::iterator it;
	
	long index = 1;
	for (gom_iter = objectMap.begin(); gom_iter != objectMap.end(); gom_iter++) {
		if (strcmp(gom_iter->first, "floor") == 0 || 
			strcmp(gom_iter->first, "room") == 0) continue;
		
		game_obj_list objects = (*gom_iter).second;
		for (it = objects.begin(); it != objects.end(); it++) {
			CEsmSubCellRef* cref = (*it)->toCREF();
			cref->SetIndex(index++, 0);
			pRecord->AddSubRecord(cref);
		}
	}
}

int Room::getPredicate(char* pred) {
	if (strcmp(pred, "center of") == 0) {
		return CENTER_OF;
	} else if (strcmp(pred, "is a") == 0) {
		return IS_A;
	} else if (strcmp(pred, "on top of") == 0) {
		return ON_TOP_OF;
	} else if (strcmp(pred, "beside") == 0) {
		return BESIDE;
	} else if (strcmp(pred, "around") == 0) {
		return AROUND;
	} else if (strcmp(pred, "edge of") == 0) {
		return EDGE_OF;
	} else if (strcmp(pred, "corner of") == 0) {
		return CORNER_OF;
	} else if (strcmp(pred, "below") == 0) {
		return BELOW;
	} else if (strcmp(pred, "inside")  == 0) {
		return INSIDE;
	}
	return UNKNOWN;
}